#! /usr/bin/env python
#
#    Generate pure C code based on definition files
#    Copyright (C) 2004  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of version 2 of the GNU General Public License as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

class Gen(object): pass
class GenClass(Gen): pass
class GenObj(Gen): pass
class GenMeth(Gen): pass
class GenFun(Gen): pass

__all__ = ['GenClass', 'GenObj', 'GenMeth', 'GenFun']

import inspect
import types

seen_modules = {}
classes = {}

def is_(type_):
    def is_type_(obj):
        return isinstance(obj, type_)
    return is_type_

def sub_(type_):
    def sub_type_(obj):
        try: issubclass(obj, int)
        except TypeError: pass
        else: return issubclass(obj, type_) and not obj is type_
    return sub_type_

class File(file):
    def __init__(self, *args, **kwargs):
        self.line = 2 # for some reason it's off by one...
        super(File, self).__init__(*args, **kwargs)

    def write(self, *data):
        data = ' '.join(data)
        self.line += len(data.split('\n'))-1
        super(File, self).write(data)

    def writeln(self, *data):
        self.write(' '.join(data) + '\n')
        

class Object(object):

    def __getitem__(self, item):
        return getattr(self, item)

    def __init__(self):
        if hasattr(self, 'obj'):
            self.docname = "NULL"
            self.doc = inspect.getdoc(self.obj)

            self.code = getattr(self, '__code__',
                    getattr(self.obj, '__code__', None))

            if not self.code:
                return

            for base in getattr(self.obj, '__bases__', []):
                if getattr(base, '__code__', None) is self.code:
                    self.code = None
                    break
            else:
                self._findcode(self.obj, self.code.split('\n'))

    def _findcode(self, obj, code):
        sourcefile = inspect.getsourcefile(obj)
        source, atline = inspect.getsourcelines(obj)
        while code[-1].strip() == '': code.pop()
        while code[0].strip() == '': code.pop(0)

        for i in range(len(source)-len(code)):
            for j in range(len(code)):
                if source[i+j].strip().replace('\\','') != code[j].strip().replace('\\',''): break
            else: break
        else:
            raise ValueError, "Couldn't find code for %s" % self.name

        self.code = code
        self.sourcefile = sourcefile
        self.sourceline = atline + i

    def _writecode(self, f):
        if self.code:
            f.writeln("#line %d \"%s\"" % (self.sourceline, self.sourcefile))
            f.writeln('\n'.join(self.code))
            f.writeln("#line %d \"%s\"" % (f.line, f.name))

    def _writedoc(self, f):
        if self.doc:
            self.docname = self.c_name + '__docstr'
            f.write("static char %s[] =" % self.docname)
            for line in self.doc.split('\n'):
                f.write("\n    \"%s\\n\"" % line)
            f.writeln(";")
        else:
            self.docname = "NULL"

class Module(Object):

    def __init__(self, name, modobj):

        self.name = name
        self.c_name = self.name
        self.obj = modobj
        self.classes = []

        self.includes = getattr(modobj, '__includes__', [])

        for classname, clsobj in inspect.getmembers(self.obj, sub_(GenClass)):
            self.classes.append(Class(self, classname, clsobj))

        added = {GenClass: True}
        remaining = {}
        for cls in self.classes:
            remaining[cls.typename] = cls

        self.classes = []
        # sort classes topologically by dependencies
        while len(remaining):
            for name, cls in remaining.items():
                for base in cls.obj.__bases__:
                    if base not in added: break
                else:
                    del remaining[name]
                    added[cls.obj] = True
                    self.classes.append(cls)
                    print cls.name

        #super(Module, self).__init__()
        Object.__init__(self)

    def write(self, f):
        self.write_prologue(f)
        self.write_body(f)
        self.write_postlogue(f)


    def write_prologue(self, f):
        f.writeln("#include <Python.h>")
        for include in self.includes:
            f.writeln("#include", include)

        self._writecode(f)

        for cls in self.classes:
            cls.write_prologue(f)


    def write_body(self, f):

        for cls in self.classes:
            cls.write_body(f)


    def write_postlogue(self, f):

        self._writedoc(f)

        f.write('\n')
        f.writeln("static PyMethodDef funcs[] = {")
        f.writeln("    { NULL }")
        f.writeln("};")

        f.write('\n')
        f.writeln("void init%s(void)" % self.name)
        f.writeln("{")
        f.writeln("    PyObject *mod = Py_InitModule3(\"%s\", funcs, %s);" % (self.name, self.docname))

        for cls in self.classes:
            cls.write_postlogue(f)
        f.writeln("}")

class Class(Object):

    def __init__(self, module, name, clsobj):
        classes[clsobj] = self
        self.module = module
        self.name = name
        self.obj = clsobj
        self.methods = []

        self.instance = self.obj()

        self.py_name = '.'.join([module.name, name])
        self.c_name = '_'.join([module.name, name])
        self.typename = 't_' + self.c_name

        self.storagevars = getattr(self.obj, '__data__', [])

        self.bases = [base for base in self.obj.__bases__ if base is not GenClass]
        # methods on the class
        for name, meth in inspect.getmembers(self.obj, is_(types.UnboundMethodType)):
            if inspect.getmodule(meth) == inspect.getmodule(self.obj):
                self.methods.append(Method(self, name, meth))

        #super(Module, self).__init__()
        Object.__init__(self)

    def write_prologue(self, f):

        self._writecode(f)

        # C level datatype
        f.write('\n')
        f.writeln("typedef struct {")
        f.writeln("    PyObject_HEAD")

        for cname, ctype in self.storagevars:
            f.writeln("    %s %s;" % (ctype, cname))

        f.writeln("} %s;" % self.c_name)

    def write_body(self, f):

        for method in self.methods:
            method.write_body(f)

        f.write('\n')
        if len(self.methods):
            self.methodsname = self.c_name + '__methods'
            f.writeln("static PyMethodDef %s[] = {" % self.methodsname)
            for method in self.methods:
                method.write_declaration(f)

            f.writeln("    { NULL }")
            f.writeln("};")

        #   type methods
        f.writeln("static PyTypeObject %s = {" % self.typename)
        f.writeln("    /* tp_head */ PyObject_HEAD_INIT(NULL)")
        f.writeln("    /* tp_internal */ 0,")
        f.writeln("    /* tp_name */ \"%s\"," % self.py_name)
        f.writeln("    /* tp_basicsize */ sizeof(%s)," % self.c_name)
        f.writeln("    /* tp_itemsize */ 0,")

        #f.writeln("    /* tp_dealloc */ (destructor) %s__dealloc_," % self.c_name)
        f.writeln("    /* tp_dealloc */ 0,")
        f.writeln("    /* tp_print */ 0,")
        f.writeln("    /* tp_getattr */ 0,")
        f.writeln("    /* tp_setattr */ 0,")
        f.writeln("    /* tp_compare */ 0,")
        f.writeln("    /* tp_repr */ 0,")

        f.writeln("    /* tp_as_number */ 0,")
        f.writeln("    /* tp_as_sequence */ 0,")
        f.writeln("    /* tp_as_mapping */ 0,")

        f.writeln("    /* tp_hash */ 0,")
        f.writeln("    /* tp_call */ 0,")
        f.writeln("    /* tp_str */ 0,")
        f.writeln("    /* tp_getattro */ (getattrofunc)PyObject_GenericGetAttr,")
        f.writeln("    /* tp_setattro */ (setattrofunc)PyObject_GenericSetAttr,")

        f.writeln("    /* tp_as_buffer */ 0,")

        f.writeln("    /* tp_flags */ Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,")

        f.writeln("    /* tp_doc */ %s," % self.docname)

        # 2.0
        f.writeln("    /* tp_traverse */ 0,")
        f.writeln("    /* tp_clear */ 0,")

        # 2.1
        f.writeln("    /* tp_richcompare */ 0,")
        f.writeln("    /* tp_weaklistoffset */ 0,")

        # 2.2
        f.writeln("    /* tp_iter */ 0,")
        f.writeln("    /* tp_iternext */ 0,")

        f.writeln("    /* tp_methods */ %s," % self.methodsname)
        f.writeln("    /* tp_members */ 0,")
        f.writeln("    /* tp_getset */ 0,")
        f.writeln("    /* tp_base */ 0,")
        f.writeln("    /* tp_dict */ 0,")
        f.writeln("    /* tp_descr_get */ 0,")
        f.writeln("    /* tp_descr_set */ 0,")
        f.writeln("    /* tp_dictoffset */ 0,")
        f.writeln("    /* tp_init */ 0,")
        f.writeln("    /* tp_alloc */ (allocfunc)PyType_GenericAlloc,")
        f.writeln("    /* tp_new */ (newfunc)PyType_GenericNew,")
        f.writeln("    /* tp_free */ 0,") #(freefunc)_PyObject_Del,")
        f.writeln("    /* tp_is_gc */ 0,")
        f.writeln("    /* tp_bases */ 0,")
        f.writeln("    /* tp_mro */ 0,")
        f.writeln("    /* tp_cache */ 0,")
        f.writeln("    /* tp_subclasses */ 0,")
        f.writeln("    /* tp_weaklist */ 0,")

        f.writeln("};")

    def write_postlogue(self, f):
        f.write('\n')
        if len(self.bases):
            f.writeln("    %s.tp_bases = Py_BuildValue(\"(%s)\", %s);" % \
                (self.typename, "O" * len(self.bases),
                ", ".join(["&"+classes[base].typename for base in self.bases])))
        f.writeln("    PyType_Ready(&%s);" % self.typename)
        f.writeln("    PyObject_SetAttrString(mod, \"%s\", (PyObject*)&%s);" % (self.name, self.typename))


class Method(Object):
                #self.methods.append(Method(self, name, meth))
    def __init__(self, cls, name, methobj):
        if name == "_init_": name = "__init__"
        self.cls = cls
        self.name = name
        self.obj = methobj

        self.py_name = '.'.join([cls.module.name, cls.name, name])
        self.c_name = '_'.join([cls.module.name, cls.name, name])

        self.argspec = inspect.getargspec(self.obj)

        if len(self.argspec[0]) > 1:
            self.parseargs = True
            self.flags = 'METH_VARARGS|METH_KEYWORDS'
        else:
            self.parseargs = False
            self.flags = 'METH_O'

        # call the function to find out more info
        self.info = self.obj(*([self.cls.instance]*len(self.argspec[0])))

        if self.parseargs:
            self.capture = []
            for arg in self.argspec[0][1:]:
                try:
                    argchar, argtype = self.info[arg]
                except KeyError:
                    raise ValueError, "Undefined Argument %r" % arg
                except TypeError:
                    print '+++++++++++++++++++++++++++'
                    print info
                    raise

                self.capture.append([argchar, argtype, arg])

            if self.argspec[1]:
                self.capture.append(['O', 'PyObject *', '*'+self.argspec[1]])
            if self.argspec[2]:
                self.capture.append(['O', 'PyObject *', '**'+self.argspec[2]])

        self.__code__ = self.info['code']

        super(Method, self).__init__()
        #Object.__init__(self)

        self.doc = self.name + inspect.formatargspec(*self.argspec) + \
                (self.doc and ('\n\n' + self.doc) or '')


    def write_declaration(self, f):
        f.writeln("    { \"%(name)s\", (PyCFunction)%(c_name)s, %(flags)s, %(docname)s }," % self)

    def write_body(self, f):
        f.write('\n')
        self._writedoc(f)

        f.writeln("static PyObject *")
        if self.parseargs:
            f.writeln("%s(%s *self, PyObject *args__, PyObject *kwargs__)" % \
                    (self.c_name, self.cls.c_name))
        else:
            f.writeln("%s(%s *self)" % (self.c_name, self.cls.c_name))
        f.writeln("{")

        if self.parseargs:
            for argchar, argtype, arg in self.capture:
                f.writeln("    %s %s = 0;" % (argtype, arg.strip('*')))
            f.write("    static char* names__[] = {")

            for argchar, argtype, arg in self.capture:
                f.write(" \"%s\"," % arg)

            f.writeln(" NULL };")

            arglist = [ c for (c,t,n) in self.capture]
            spec = self.argspec
            opt = len(spec[3] or []) + bool(spec[1]) + bool(spec[2])
            if opt > 1:
                arglist.insert(len(self.argspec[0])-opt+1, '|')
            arglist.append(':' + self.name)
            arglist = ''.join(arglist)

            f.write("    if (!PyArg_ParseTupleAndKeywords(args__, kwargs__, \"%s\", names__"%arglist)
            for c, t, n in self.capture:
                f.write(", &%s" % n.strip('*'))
            f.writeln("))")
            f.writeln("    {")
            f.writeln("        return NULL;")
            f.writeln("    }")


        self._writecode(f)

        if 'return' not in self.code[-1]:
            f.writeln("    Py_INCREF(Py_None); return Py_None;")

        f.writeln("}")
        f.write('\n')


